@@ -44,8 +44,9 @@ class ApplicationMenuItem extends PopupMenu.PopupBaseMenuItem {
         this._app = app;
         this._button = button;
 
-        this._iconBin = new St.Bin();
-        this.add_child(this._iconBin);
+        this._icon = this.getDragActor();
+        this._icon.style_class = 'icon-dropshadow';
+        this.add_child(this._icon);
 
         let appLabel = new St.Label({
             text: app.get_name(),
@@ -55,11 +56,6 @@ class ApplicationMenuItem extends PopupMenu.PopupBaseMenuItem {
         this.add_child(appLabel);
         this.label_actor = appLabel;
 
-        let textureCache = St.TextureCache.get_default();
-        textureCache.connectObject('icon-theme-changed',
-            () => this._updateIcon(), this);
-        this._updateIcon();
-
         this._delegate = this;
         let draggable = DND.makeDraggable(this);
 
@@ -69,6 +65,8 @@ class ApplicationMenuItem extends PopupMenu.PopupBaseMenuItem {
                 return maybeStartDrag.call(draggable, event);
             return false;
         };
+
+        this.connect('notify::active', this._onActiveChanged.bind(this));
     }
 
     activate(event) {
@@ -80,10 +78,11 @@ class ApplicationMenuItem extends PopupMenu.PopupBaseMenuItem {
         Main.overview.hide();
     }
 
-    setActive(active, params) {
-        if (active)
-            this._button.scrollToButton(this);
-        super.setActive(active, params);
+    _onActiveChanged() {
+        if (!this.active)
+            return;
+
+        this._button.scrollToButton(this);
     }
 
     setDragEnabled(enabled) {
@@ -95,13 +94,7 @@ class ApplicationMenuItem extends PopupMenu.PopupBaseMenuItem {
     }
 
     getDragActorSource() {
-        return this._iconBin;
-    }
-
-    _updateIcon() {
-        let icon = this.getDragActor();
-        icon.style_class = 'icon-dropshadow';
-        this._iconBin.set_child(icon);
+        return this._icon;
     }
 }
 
@@ -495,7 +488,7 @@ class ApplicationsButton extends PanelMenu.Button {
                 let id;
                 try {
                     id = entry.get_desktop_file_id(); // catch non-UTF8 filenames
-                } catch {
+                } catch (e) {
                     continue;
                 }
                 let app = appSys.lookup_app(id);
@@ -547,7 +540,9 @@ class ApplicationsButton extends PanelMenu.Button {
         let section = new PopupMenu.PopupMenuSection();
         this.menu.addMenuItem(section);
         this.mainBox = new St.BoxLayout({layoutManager: new MainLayout()});
-        this.leftBox = new St.BoxLayout({vertical: true});
+        this.leftBox = new St.BoxLayout({
+            orientation: Clutter.Orientation.VERTICAL,
+        });
         this.applicationsScrollBox = new St.ScrollView({
             style_class: 'apps-menu vfade',
             x_expand: true,
@@ -557,9 +552,13 @@ class ApplicationsButton extends PanelMenu.Button {
         });
         this.leftBox.add_child(this.categoriesScrollBox);
 
-        this.applicationsBox = new St.BoxLayout({vertical: true});
+        this.applicationsBox = new St.BoxLayout({
+            orientation: Clutter.Orientation.VERTICAL,
+        });
         this.applicationsScrollBox.set_child(this.applicationsBox);
-        this.categoriesBox = new St.BoxLayout({vertical: true});
+        this.categoriesBox = new St.BoxLayout({
+            orientation: Clutter.Orientation.VERTICAL,
+        });
         this.categoriesScrollBox.set_child(this.categoriesBox);
 
         this.mainBox.add_child(this.leftBox);
