@@ -26,12 +26,37 @@ import {WorkspaceIndicator} from './workspaceIndicator.js';
 const ICON_TEXTURE_SIZE = 24;
 const DND_ACTIVATE_TIMEOUT = 500;
 
+const MIN_DRAG_UPDATE_INTERVAL = 500 * GLib.TIME_SPAN_MILLISECOND;
+
+const DRAG_OPACITY = 0.3;
+const DRAG_FADE_DURATION = 200;
+
+const DRAG_RESIZE_DURATION = 400;
+
+const DRAG_PROXIMITY_THRESHOLD = 30;
+
+const SAVED_POSITIONS_KEY = 'window-list-positions';
+
+const ATTENTION_INDICATOR_MAX_SCALE = 0.4;
+const ATTENTION_INDICATOR_TRANSITION_DURATION = 300;
+
 const GroupingMode = {
     NEVER: 0,
     AUTO: 1,
     ALWAYS: 2,
 };
 
+class DragPlaceholderItem extends DashItemContainer {
+    static {
+        GObject.registerClass(this);
+    }
+
+    constructor() {
+        super();
+        this.setChild(new St.Bin({style_class: 'placeholder'}));
+    }
+}
+
 /**
  * @param {Shell.App} app - an app
  * @returns {number} - the smallest stable sequence of the app's windows
@@ -105,26 +130,89 @@ class WindowContextMenu extends PopupMenu.PopupMenu {
     }
 }
 
-class WindowTitle extends St.BoxLayout {
+class TitleWidget extends St.Widget {
+    static {
+        GObject.registerClass({
+            GTypeFlags: GObject.TypeFlags.ABSTRACT,
+            Properties: {
+                'abstract-label': GObject.ParamSpec.boolean(
+                    'abstract-label', null, null,
+                    GObject.ParamFlags.READWRITE,
+                    false),
+            },
+        }, this);
+    }
+
+    constructor() {
+        super({
+            layout_manager: new Clutter.BinLayout(),
+            x_expand: true,
+            y_expand: true,
+        });
+
+        const hbox = new St.BoxLayout({
+            style_class: 'window-button-box',
+            x_expand: true,
+            y_expand: true,
+        });
+        this.add_child(hbox);
+
+        this._icon = new St.Bin({
+            style_class: 'window-button-icon',
+        });
+        hbox.add_child(this._icon);
+
+        this._label = new St.Label({
+            y_align: Clutter.ActorAlign.CENTER,
+        });
+        hbox.add_child(this._label);
+        this.label_actor = this._label;
+
+        this.bind_property('abstract-label',
+            this._label, 'visible',
+            GObject.BindingFlags.SYNC_CREATE |
+            GObject.BindingFlags.INVERT_BOOLEAN);
+
+        this._abstractLabel = new St.Widget({
+            style_class: 'window-button-abstract-label',
+            x_expand: true,
+            y_expand: true,
+        });
+        hbox.add_child(this._abstractLabel);
+
+        this.bind_property('abstract-label',
+            this._abstractLabel, 'visible',
+            GObject.BindingFlags.SYNC_CREATE);
+
+        this._attentionIndicator = new St.Widget({
+            style_class: 'window-button-attention-indicator',
+            x_expand: true,
+            y_expand: true,
+            y_align: Clutter.ActorAlign.END,
+            scale_x: 0,
+        });
+        this._attentionIndicator.set_pivot_point(0.5, 0.5);
+        this.add_child(this._attentionIndicator);
+    }
+
+    setNeedsAttention(enable) {
+        this._attentionIndicator.ease({
+            scaleX: enable ? ATTENTION_INDICATOR_MAX_SCALE : 0,
+            duration: ATTENTION_INDICATOR_TRANSITION_DURATION,
+        });
+    }
+}
+
+class WindowTitle extends TitleWidget {
     static {
         GObject.registerClass(this);
     }
 
     constructor(metaWindow) {
-        super({
-            style_class: 'window-button-box',
-            x_expand: true,
-            y_expand: true,
-        });
+        super();
 
         this._metaWindow = metaWindow;
 
-        this._icon = new St.Bin({style_class: 'window-button-icon'});
-        this.add_child(this._icon);
-        this.label_actor = new St.Label({y_align: Clutter.ActorAlign.CENTER});
-        this.label_actor.clutter_text.single_line_mode = true;
-        this.add_child(this.label_actor);
-
         this._metaWindow.connectObject(
             'notify::wm-class',
             () => this._updateIcon(), GObject.ConnectFlags.AFTER,
@@ -132,10 +220,13 @@ class WindowTitle extends St.BoxLayout {
             () => this._updateIcon(), GObject.ConnectFlags.AFTER,
             'notify::title', () => this._updateTitle(),
             'notify::minimized', () => this._minimizedChanged(),
+            'notify::demands-attention', () => this._updateNeedsAttention(),
+            'notify::urgent', () => this._updateNeedsAttention(),
             this);
 
         this._updateIcon();
         this._minimizedChanged();
+        this._updateNeedsAttention();
     }
 
     _minimizedChanged() {
@@ -143,14 +234,19 @@ class WindowTitle extends St.BoxLayout {
         this._updateTitle();
     }
 
+    _updateNeedsAttention() {
+        const {urgent, demandsAttention} = this._metaWindow;
+        this.setNeedsAttention(urgent || demandsAttention);
+    }
+
     _updateTitle() {
         if (!this._metaWindow.title)
             return;
 
         if (this._metaWindow.minimized)
-            this.label_actor.text = '[%s]'.format(this._metaWindow.title);
+            this._label.text = '[%s]'.format(this._metaWindow.title);
         else
-            this.label_actor.text = this._metaWindow.title;
+            this._label.text = this._metaWindow.title;
     }
 
     _updateIcon() {
@@ -166,6 +262,100 @@ class WindowTitle extends St.BoxLayout {
     }
 }
 
+class AppTitle extends TitleWidget {
+    static {
+        GObject.registerClass(this);
+    }
+
+    constructor(app) {
+        super();
+
+        this._app = app;
+        this._windows = new Set();
+
+        this._icon.child = app.create_icon_texture(ICON_TEXTURE_SIZE);
+        this._label.text = app.get_name();
+
+        this._app.connectObject(
+            'windows-changed', () => this._onWindowsChanged(),
+            this);
+        this._onWindowsChanged();
+
+        this.connect('destroy', () => {
+            console.debug(`Clearing windows of app ${this._app.id}`);
+            this._windows.clear();
+        });
+    }
+
+    _onWindowsChanged() {
+        const windows = this._app.get_windows();
+        const removed = [...this._windows].filter(w => !windows.includes(w));
+        removed.forEach(w => this._untrackWindow(w));
+        windows.forEach(w => this._trackWindow(w));
+        this._updateNeedsAttention();
+    }
+
+    _trackWindow(window) {
+        if (this._windows.has(window))
+            return;
+
+        console.debug(`Tracking window ${window} for app ${this._app.id}`);
+        window.connectObject(
+            'notify::urgent', () => this._updateNeedsAttention(),
+            'notify::demands-attention', () => this._updateNeedsAttention(),
+            this);
+        this._windows.add(window);
+    }
+
+    _untrackWindow(window) {
+        if (!this._windows.delete(window))
+            return;
+
+        console.debug(`Untracking window ${window} for app ${this._app.id}`);
+        window.disconnectObject(this);
+    }
+
+    _updateNeedsAttention() {
+        const needsAttention =
+            [...this._windows].some(w => w.urgent || w.demandsAttention);
+        this.setNeedsAttention(needsAttention);
+    }
+}
+
+class DragActor extends St.Bin {
+    static {
+        GObject.registerClass(this);
+    }
+
+    constructor(source, titleActor) {
+        super({
+            style_class: 'window-button-drag-actor',
+            child: titleActor,
+            width: source.width,
+        });
+
+        this.source = source;
+    }
+
+    setTargetWidth(width) {
+        const currentWidth = this.width;
+
+        // set width immediately so shell's DND code uses correct values
+        this.set({width});
+
+        // then transition from the original to the new width
+        const laters = global.compositor.get_laters();
+        laters.add(Meta.LaterType.BEFORE_REDRAW, () => {
+            this.set({width: currentWidth});
+            this.ease({
+                width,
+                duration: DRAG_RESIZE_DURATION,
+            });
+            return GLib.SOURCE_REMOVE;
+        });
+    }
+}
+
 class BaseButton extends DashItemContainer {
     static {
         GObject.registerClass({
@@ -176,6 +366,10 @@ class BaseButton extends DashItemContainer {
                     GObject.ParamFlags.READWRITE,
                     false),
             },
+            Signals: {
+                'drag-begin': {},
+                'drag-end': {},
+            },
         }, this);
     }
 
@@ -220,16 +414,30 @@ class BaseButton extends DashItemContainer {
                 this._windowEnteredOrLeftMonitor.bind(this),
                 this);
         }
+
+        this._button._delegate = this;
+        this._draggable = DND.makeDraggable(this._button);
+        this._draggable.connect('drag-begin', () => {
+            this._removeLongPressTimeout();
+            this.emit('drag-begin');
+        });
+        this._draggable.connect('drag-cancelled', () => {
+            this._draggable._dragActor?.setTargetWidth(this.width);
+            this.emit('drag-end');
+        });
+        this._draggable.connect('drag-end', () => this.emit('drag-end'));
     }
 
     get active() {
         return this._button.has_style_class_name('focused');
     }
 
+    // eslint-disable-next-line camelcase
     get ignore_workspace() {
         return this._ignoreWorkspace;
     }
 
+    // eslint-disable-next-line camelcase
     set ignore_workspace(ignore) {
         if (this._ignoreWorkspace === ignore)
             return;
@@ -301,6 +509,28 @@ class BaseButton extends DashItemContainer {
         this._onClicked(this, 1);
     }
 
+    getDragActor() {
+        const titleActor = this._createTitleActor();
+        titleActor.set({abstractLabel: true});
+
+        const dragActor = new DragActor(this, titleActor);
+
+        const [, natWidth] = this.get_preferred_width(-1);
+        const targetWidth = Math.min(natWidth / 2, this.width);
+        dragActor.setTargetWidth(targetWidth);
+
+        return dragActor;
+    }
+
+    getDragActorSource() {
+        return this;
+    }
+
+    _createTitleActor() {
+        throw new GObject.NotImplementedError(
+            `_createTitleActor in ${this.constructor.name}`);
+    }
+
     _onClicked(_actor, _button) {
         throw new GObject.NotImplementedError(
             `_onClicked in ${this.constructor.name}`);
@@ -411,7 +641,7 @@ class WindowButton extends BaseButton {
 
         this._updateVisibility();
 
-        this._windowTitle = new WindowTitle(this.metaWindow);
+        this._windowTitle = this._createTitleActor();
         this._button.set_child(this._windowTitle);
         this.label_actor = this._windowTitle.label_actor;
 
@@ -427,6 +657,14 @@ class WindowButton extends BaseButton {
         this._updateStyle();
     }
 
+    get id() {
+        return `window:${this.metaWindow.get_id()}`;
+    }
+
+    _createTitleActor() {
+        return new WindowTitle(this.metaWindow);
+    }
+
     _onClicked(actor, button) {
         if (this._contextMenu.isOpen) {
             this._contextMenu.close();
@@ -543,33 +781,6 @@ class AppButton extends BaseButton {
         this.app = app;
         this._updateVisibility();
 
-        let stack = new St.Widget({layout_manager: new Clutter.BinLayout()});
-        this._button.set_child(stack);
-
-        this._singleWindowTitle = new St.Bin({
-            x_expand: true,
-        });
-        stack.add_child(this._singleWindowTitle);
-
-        this._multiWindowTitle = new St.BoxLayout({
-            style_class: 'window-button-box',
-            x_expand: true,
-        });
-        stack.add_child(this._multiWindowTitle);
-
-        this._icon = new St.Bin({
-            style_class: 'window-button-icon',
-            child: app.create_icon_texture(ICON_TEXTURE_SIZE),
-        });
-        this._multiWindowTitle.add_child(this._icon);
-
-        let label = new St.Label({
-            text: app.get_name(),
-            y_align: Clutter.ActorAlign.CENTER,
-        });
-        this._multiWindowTitle.add_child(label);
-        this._multiWindowTitle.label_actor = label;
-
         this._menuManager = new PopupMenu.PopupMenuManager(this);
         this._menu = new PopupMenu.PopupMenu(this, 0.5, St.Side.BOTTOM);
         this._menu.connect('open-state-changed',
@@ -579,12 +790,6 @@ class AppButton extends BaseButton {
         this._menuManager.addMenu(this._menu);
         Main.uiGroup.add_child(this._menu.actor);
 
-        this._appContextMenu = new AppContextMenu(this);
-        this._appContextMenu.connect('open-state-changed',
-            this._onMenuStateChanged.bind(this));
-        this._appContextMenu.actor.hide();
-        Main.uiGroup.add_child(this._appContextMenu.actor);
-
         this.app.connectObject('windows-changed',
             () => this._windowsChanged(), this);
         this._windowsChanged();
@@ -595,6 +800,10 @@ class AppButton extends BaseButton {
         this._updateStyle();
     }
 
+    get id() {
+        return `app:${this.app.get_id()}`;
+    }
+
     _windowEnteredOrLeftMonitor(metaDisplay, monitorIndex, metaWindow) {
         if (this._windowTracker.get_window_app(metaWindow) === this.app &&
             monitorIndex === this._monitorIndex) {
@@ -631,36 +840,40 @@ class AppButton extends BaseButton {
     }
 
     _windowsChanged() {
-        let windows = this.getWindowList();
-        this._singleWindowTitle.visible = windows.length === 1;
-        this._multiWindowTitle.visible = !this._singleWindowTitle.visible;
+        const windows = this.getWindowList();
+        const singleWindowMode = windows.length === 1;
 
-        if (this._singleWindowTitle.visible) {
-            if (!this._windowTitle) {
-                this.metaWindow = windows[0];
-                this._windowTitle = new WindowTitle(this.metaWindow);
-                this._singleWindowTitle.child = this._windowTitle;
-                this._windowContextMenu = new WindowContextMenu(this, this.metaWindow);
-                this._windowContextMenu.connect(
-                    'open-state-changed', this._onMenuStateChanged.bind(this));
-                Main.uiGroup.add_child(this._windowContextMenu.actor);
-                this._windowContextMenu.actor.hide();
-                this._contextMenuManager.addMenu(this._windowContextMenu);
-            }
-            this._contextMenuManager.removeMenu(this._appContextMenu);
-            this._contextMenu = this._windowContextMenu;
-            this.label_actor = this._windowTitle.label_actor;
+        if (this._singleWindowMode === singleWindowMode)
+            return;
+
+        this._singleWindowMode = singleWindowMode;
+
+        this._button.child?.destroy();
+        this._contextMenu?.destroy();
+
+        if (this._singleWindowMode) {
+            const [window] = windows;
+            this._contextMenu = new WindowContextMenu(this, window);
         } else {
-            if (this._windowTitle) {
-                this.metaWindow = null;
-                this._singleWindowTitle.child = null;
-                this._windowTitle = null;
-                this._windowContextMenu.destroy();
-                this._windowContextMenu = null;
-            }
-            this._contextMenu = this._appContextMenu;
-            this._contextMenuManager.addMenu(this._appContextMenu);
-            this.label_actor = this._multiWindowTitle.label_actor;
+            this._contextMenu = new AppContextMenu(this);
+        }
+
+        this._button.child = this._createTitleActor();
+        this.label_actor = this._button.child.label_actor;
+
+        this._contextMenu.connect(
+            'open-state-changed', this._onMenuStateChanged.bind(this));
+        Main.uiGroup.add_child(this._contextMenu.actor);
+        this._contextMenu.actor.hide();
+        this._contextMenuManager.addMenu(this._contextMenu);
+    }
+
+    _createTitleActor() {
+        if (this._singleWindowMode) {
+            const [window] = this.getWindowList();
+            return new WindowTitle(window);
+        } else {
+            return new AppTitle(this.app);
         }
     }
 
@@ -770,20 +983,11 @@ class WindowList extends St.Widget {
             () => this._onWorkspaceMenuSet(), this);
         this._onWorkspaceMenuSet();
 
-        const inOverview = Main.overview.visible ||
-            (Main.layoutManager._startingUp && Main.sessionMode.hasOverview);
-
-        const overviewChromeOptions = {
-            affectsStruts: true,
-        };
         const chromeOptions = {
-            ...overviewChromeOptions,
+            affectsStruts: true,
             trackFullscreen: true,
         };
-        Main.layoutManager.addChrome(this, inOverview
-            ? overviewChromeOptions
-            : chromeOptions);
-
+        Main.layoutManager.addChrome(this, chromeOptions);
         Main.uiGroup.set_child_above_sibling(this, Main.layoutManager.panelBox);
         Main.ctrlAltTabManager.addGroup(this, _('Window List'), 'start-here-symbolic');
 
@@ -820,17 +1024,20 @@ class WindowList extends St.Widget {
 
         Main.overview.connectObject(
             'showing', () => {
-                this._retrackChrome(overviewChromeOptions);
+                Main.layoutManager.untrackChrome(this);
                 this.hide();
                 this._updateKeyboardAnchor();
             },
             'hidden', () => {
-                this._retrackChrome(chromeOptions);
+                Main.layoutManager.trackChrome(this, chromeOptions);
                 this.visible = !this._monitor.inFullscreen;
                 this._updateKeyboardAnchor();
             }, this);
 
         global.display.connectObject('in-fullscreen-changed', () => {
+            // Work-around for initial change from unknown to !fullscreen
+            if (Main.overview.visible)
+                this.hide();
             this._updateKeyboardAnchor();
         }, this);
 
@@ -839,17 +1046,28 @@ class WindowList extends St.Widget {
             'window-created', (dsp, win) => this._addWindow(win, true), this);
 
         Main.xdndHandler.connectObject(
-            'drag-begin', () => this._monitorDrag(),
-            'drag-end', () => this._stopMonitoringDrag(),
+            'drag-begin', () => this._monitorXdndDrag(),
+            'drag-end', () => this._stopMonitoringXdndDrag(),
             this);
 
-        this._dragMonitor = {
-            dragMotion: this._onDragMotion.bind(this),
+        this._xdndDragMonitor = {
+            dragMotion: this._onXdndDragMotion.bind(this),
+        };
+
+        this._itemDragMonitor = {
+            dragMotion: this._onItemDragMotion.bind(this),
+            dragDrop: this._onItemDragDrop.bind(this),
         };
 
         this._dndTimeoutId = 0;
         this._dndWindow = null;
 
+        this._dragPlaceholder = null;
+        this._dragPlaceholderPos = -1;
+        this._lastPlaceholderUpdate = 0;
+
+        this._delegate = this;
+
         this._settings = settings;
         this._settings.connectObject('changed::grouping-mode',
             () => this._groupingModeChanged(), this);
@@ -885,11 +1103,6 @@ class WindowList extends St.Widget {
             this._monitor.y + this._monitor.height - this.height);
     }
 
-    _retrackChrome(options) {
-        Main.layoutManager.untrackChrome(this);
-        Main.layoutManager.trackChrome(this, options);
-    }
-
     _updateWorkspaceIndicatorVisibility() {
         let workspaceManager = global.workspace_manager;
         let hasWorkspaces = this._mutterSettings.get_boolean('dynamic-workspaces') ||
@@ -967,6 +1180,8 @@ class WindowList extends St.Widget {
             for (let i = 0; i < apps.length; i++)
                 this._addApp(apps[i], false);
         }
+
+        this._restorePositions();
     }
 
     _updateKeyboardAnchor() {
@@ -984,14 +1199,37 @@ class WindowList extends St.Widget {
             this._removeApp(app);
     }
 
-    _addApp(app, animate) {
-        let button = new AppButton(app, this._perMonitor, this._monitor.index);
+    _addButton(button, animate) {
         this._settings.bind('display-all-workspaces',
             button, 'ignore-workspace', Gio.SettingsBindFlags.GET);
+
+        button.connect('drag-begin', () => {
+            button.ease({
+                opacity: 255 * DRAG_OPACITY,
+                duration: DRAG_FADE_DURATION,
+            });
+
+            this._monitorItemDrag();
+        });
+        button.connect('drag-end', () => {
+            button.ease({
+                opacity: 255,
+                duration: DRAG_FADE_DURATION,
+            });
+
+            this._stopMonitoringItemDrag();
+            this._clearDragPlaceholder();
+        });
+
         this._windowList.add_child(button);
         button.show(animate);
     }
 
+    _addApp(app, animate) {
+        const button = new AppButton(app, this._perMonitor, this._monitor.index);
+        this._addButton(button, animate);
+    }
+
     _removeApp(app) {
         let children = this._windowList.get_children();
         let child = children.find(c => c.app === app);
@@ -1012,11 +1250,8 @@ class WindowList extends St.Widget {
         this._windowSignals.set(
             win, win.connect('unmanaged', () => this._removeWindow(win)));
 
-        let button = new WindowButton(win, this._perMonitor, this._monitor.index);
-        this._settings.bind('display-all-workspaces',
-            button, 'ignore-workspace', Gio.SettingsBindFlags.GET);
-        this._windowList.add_child(button);
-        button.show(animate);
+        const button = new WindowButton(win, this._perMonitor, this._monitor.index);
+        this._addButton(button, animate);
     }
 
     _removeWindow(win) {
@@ -1036,16 +1271,135 @@ class WindowList extends St.Widget {
         child?.animateOutAndDestroy();
     }
 
-    _monitorDrag() {
-        DND.addDragMonitor(this._dragMonitor);
+    _clearDragPlaceholder() {
+        this._dragPlaceholder?.animateOutAndDestroy();
+        this._dragPlaceholder = null;
+        this._dragPlaceholderPos = -1;
     }
 
-    _stopMonitoringDrag() {
-        DND.removeDragMonitor(this._dragMonitor);
+    handleDragOver(source, _actor, x, _y, _time) {
+        if (!(source instanceof BaseButton))
+            return DND.DragMotionResult.NO_DROP;
+
+        const buttons = this._windowList.get_children().filter(c => c instanceof BaseButton);
+        const buttonPos = buttons.indexOf(source);
+        const numButtons = buttons.length;
+        let boxWidth = this._windowList.width;
+
+        // Transform to window list coordinates for index calculation
+        // (mostly relevant for RTL to discard workspace indicator etc.)
+        x -= this._windowList.x;
+
+        const rtl = this.text_direction === Clutter.TextDirection.RTL;
+        let pos = rtl
+            ? numButtons - Math.round(x * numButtons / boxWidth)
+            : Math.round(x * numButtons / boxWidth);
+
+        pos = Math.clamp(pos, 0, numButtons);
+
+        const timeDelta =
+            GLib.get_monotonic_time() - this._lastPlaceholderUpdate;
+
+        if (pos !== this._dragPlaceholderPos && timeDelta >= MIN_DRAG_UPDATE_INTERVAL) {
+            this._clearDragPlaceholder();
+            this._dragPlaceholderPos = pos;
+
+            this._lastPlaceholderUpdate = GLib.get_monotonic_time();
+
+            // Don't allow positioning before or after self
+            if (pos === buttonPos || pos === buttonPos + 1)
+                return DND.DragMotionResult.CONTINUE;
+
+            this._dragPlaceholder = new DragPlaceholderItem();
+            const sibling = buttons[pos] ?? null;
+            if (sibling)
+                this._windowList.insert_child_below(this._dragPlaceholder, sibling);
+            else
+                this._windowList.insert_child_above(this._dragPlaceholder, null);
+            this._dragPlaceholder.show(true);
+        }
+
+        return this._dragPlaceholder
+            ? DND.DragMotionResult.MOVE_DROP
+            : DND.DragMotionResult.NO_DROP;
+    }
+
+    acceptDrop(source, _actor, _x, _y, _time) {
+        if (this._dragPlaceholderPos >= 0)
+            this._windowList.set_child_at_index(source, this._dragPlaceholderPos);
+
+        this._clearDragPlaceholder();
+
+        this._savePositions();
+
+        return true;
+    }
+
+    _getPositionStateKey() {
+        return `${SAVED_POSITIONS_KEY}:${this._monitor.index}`;
+    }
+
+    _savePositions() {
+        const buttons = this._windowList.get_children()
+            .filter(b => b instanceof BaseButton);
+        global.set_runtime_state(this._getPositionStateKey(),
+            new GLib.Variant('as', buttons.map(b => b.id)));
+    }
+
+    _restorePositions() {
+        const positions = global.get_runtime_state('as',
+            this._getPositionStateKey())?.deepUnpack() ?? [];
+
+        for (const button of this._windowList.get_children()) {
+            const pos = positions.indexOf(button.id);
+            if (pos > -1)
+                this._windowList.set_child_at_index(button, pos);
+        }
+    }
+
+    _monitorItemDrag() {
+        DND.addDragMonitor(this._itemDragMonitor);
+    }
+
+    _stopMonitoringItemDrag() {
+        DND.removeDragMonitor(this._itemDragMonitor);
+    }
+
+    _onItemDragMotion(dragEvent) {
+        const {source, targetActor, dragActor, x, y} = dragEvent;
+
+        const hasTarget = this._windowList.contains(targetActor);
+        const isNear = Math.abs(y - this.y) < DRAG_PROXIMITY_THRESHOLD;
+
+        if (hasTarget || isNear)
+            return this.handleDragOver(source, dragActor, x, y);
+
+        this._clearDragPlaceholder();
+        return DND.DragMotionResult.CONTINUE;
+    }
+
+    _onItemDragDrop(dropEvent) {
+        if (this._dragPlaceholderPos < 0)
+            return DND.DragDropResult.CONTINUE;
+
+        const {source} = dropEvent.dropActor;
+        this.acceptDrop(source);
+        dropEvent.dropActor.destroy();
+        // HACK: SUCESS would make more sense, but results in gnome-shell
+        // skipping all drag-end code
+        return DND.DragDropResult.CONTINUE;
+    }
+
+    _monitorXdndDrag() {
+        DND.addDragMonitor(this._xdndDragMonitor);
+    }
+
+    _stopMonitoringXdndDrag() {
+        DND.removeDragMonitor(this._xdndDragMonitor);
         this._removeActivateTimeout();
     }
 
-    _onDragMotion(dragEvent) {
+    _onXdndDragMotion(dragEvent) {
         if (Main.overview.visible ||
             !this.contains(dragEvent.targetActor)) {
             this._removeActivateTimeout();
@@ -1093,7 +1447,7 @@ class WindowList extends St.Widget {
         this._windowSignals.forEach((id, win) => win.disconnect(id));
         this._windowSignals.clear();
 
-        this._stopMonitoringDrag();
+        this._stopMonitoringXdndDrag();
 
         this._settings.disconnectObject();
         this._settings = null;
