@@ -132,7 +132,7 @@ class WorkspaceThumbnail extends St.Button {
         const box = new St.BoxLayout({
             style_class: 'workspace-box',
             y_expand: true,
-            vertical: true,
+            orientation: Clutter.Orientation.VERTICAL,
         });
         this.set_child(box);
 
@@ -459,6 +459,8 @@ export class WorkspaceIndicator extends PanelMenu.Button {
         this._thumbnails = new WorkspacePreviews();
         container.add_child(this._thumbnails);
 
+        this._workspacesItems = [];
+
         workspaceManager.connectObject(
             'workspace-switched', this._onWorkspaceSwitched.bind(this), GObject.ConnectFlags.AFTER,
             this);
